use crate::{schema, user::UserExt, UIDCError};
use microrm::{cli::CLIError, prelude::*, schema::entity::EntityID};
use ring::rand::SecureRandom;

#[derive(Debug)]
pub struct UserInterface;

#[derive(Debug, clap::Subcommand)]
pub enum UserCommands {
    Create {
        username: String,
    },
    UpdateAuth {
        username: String,
        #[clap(short = 'p', long, action = clap::ArgAction::Count)]
        password: u8,
        #[clap(short = 't', long, action = clap::ArgAction::Count)]
        totp: u8,
    },
    RegisterExternalAuth {
        username: String,
        provider: schema::ExternalAuthProvider,
    },
    OnetimeLink {
        username: String,
        #[clap(long)]
        endpoint: Option<String>,
        #[clap(long)]
        expiry: Option<String>,
    },
}

impl microrm::cli::EntityInterface for UserInterface {
    type Error = UIDCError;
    type Entity = schema::User;
    type Context = super::RunArgs; // microrm::schema::Stored<schema::Realm>;
    type CustomCommand = UserCommands;

    fn run_custom(
        ctx: &Self::Context,
        cmd: Self::CustomCommand,
        query_ctx: impl Queryable<EntityOutput = Self::Entity> + Insertable<Self::Entity>,
    ) -> Result<(), Self::Error> {
        match cmd {
            UserCommands::Create { username } => {
                query_ctx.insert(schema::User {
                    realm: ctx.realm.id(),
                    username,
                    pending_external_auths: Default::default(),
                    auth: Default::default(),
                    groups: Default::default(),
                })?;
            }
            UserCommands::UpdateAuth {
                username,
                password,
                totp,
            } => {
                let user = query_ctx
                    .with(schema::User::Username, &username)
                    .first()
                    .get()?
                    .ok_or(Self::Error::no_such_entity("user", username))?;

                if password == 0 && totp == 0 {
                    println!("No changes requested!");
                    return Ok(());
                }
                if password > 0 {
                    let raw_pass = rpassword::prompt_password("Enter new user password: ").unwrap();
                    user.set_new_password(raw_pass.as_bytes())?;
                }
                if totp > 0 {
                    let (new_secret, new_uri) = user.generate_totp_with_uri()?;
                    println!("Please confirm you can generate tokens with the new secret:");
                    qr2term::print_qr(new_uri.as_str())
                        .map_err(|_| UIDCError::Abort("could not display QR code"))?;
                    let new_challenge = schema::AuthChallenge {
                        user_id: user.id(),
                        challenge_type: schema::AuthChallengeType::Totp.into(),
                        public: vec![],
                        secret: new_secret.clone(),
                        enabled: true,
                    };

                    loop {
                        let digits = rpassword::prompt_password("TOTP code: ").unwrap();
                        if new_challenge.verify_totp_challenge(digits.as_bytes())? {
                            break;
                        }
                    }
                    user.set_new_totp(new_secret.as_slice())?;
                }
            }
            UserCommands::RegisterExternalAuth { username, provider } => {
                let mut user = query_ctx
                    .with(schema::User::Username, &username)
                    .first()
                    .get()?
                    .ok_or(Self::Error::no_such_entity("user", username))?;

                user.pending_external_auths.as_mut().push(provider);
                user.pending_external_auths.as_mut().sort();
                user.pending_external_auths.as_mut().dedup();
                user.sync().expect("couldn't sync user model");
            }
            UserCommands::OnetimeLink {
                username,
                endpoint,
                expiry,
            } => {
                let user = query_ctx
                    .with(schema::User::Username, &username)
                    .first()
                    .get_ids()?
                    .ok_or(Self::Error::no_such_entity("user", username))?;

                let rng = ring::rand::SystemRandom::new();
                let mut code = [0u8; 16];
                rng.fill(&mut code)
                    .map_err(|_| UIDCError::Abort("couldn't generate random values"))?;

                let code = base64::encode_config(code, base64::URL_SAFE_NO_PAD);

                let expiry = time::OffsetDateTime::now_utc()
                    .checked_add(time::Duration::days(7))
                    .unwrap();

                ctx.realm
                    .single_use_auth
                    .insert(schema::SingleUseAuth {
                        code: code.clone(),
                        user,
                        expiry,
                    })
                    .expect("couldn't insert single-use authentication code?");

                let target = tide::http::Url::parse_with_params(
                    format!(
                        "{base}/{realm}/v1/session/onetime",
                        base = ctx.config.base_url,
                        realm = ctx.realm.shortname
                    )
                    .as_str(),
                    &[
                        ("code", code.as_str()),
                        ("redirect", endpoint.as_deref().unwrap_or("")),
                    ],
                )
                .expect("couldn't generate URL");

                println!("target URL: {target}");
            }
        }

        Ok(())
    }

    fn should_override(
        _entity: &'static str,
        field: &'static str,
        _role: microrm::cli::ValueRole,
    ) -> bool {
        field == "realm"
    }

    fn override_for(
        ctx: &Self::Context,
        _entity: &'static str,
        field: &'static str,
        _role: microrm::cli::ValueRole,
    ) -> String {
        if field == "realm" {
            format!("{}", ctx.realm.id().into_raw())
        } else {
            unreachable!()
        }
    }
}
